/*
Copyright, 1994-1995, The Regents of the University of California. 
This software was produced under U.S. Government contract(XXXXX)
by Los Alamos National Laboratory, which is operated by the University 
of California for the U.S. Department of Energy. The U.S Government 
is licensed to use, reproduce, and distribute this software. Permission 
is granted to the public to copy and use this software without charge, 
provided that this Notice and any statement of authorship are reproduced 
on all copies. Neither the Government nor the University makes any 
warranty, express or implied, or assumes any liability or responsibility 
for the use of this software.
*/

/* $Id: calc.c,v 1.2 96/09/11 13:32:54 arve Exp $ */

#include <ctype.h>
//++pac
#ifndef _PLAN9
#include <malloc.h>
#endif
#include <math.h>
#include <stdio.h>

#include "matrix.h"
#include "seqlist.h"
#include "structures.h"
#include "struct_methods.h"
#include "calc.h"


/* #define CALC_DEBUG */

/* 
   calc_free_matrix_list
   Frees everything in a list of data except the 
   sequence name strings.
 */
void calc_free_matrix_list(matrix_list *ml)
{
    matrix_list *tmp;
    
    while (ml) {
	tmp = ml->next;
	matrix_free(ml->change_count);
	matrix_free(ml->probability_estimate);
	free(ml);
	ml = tmp;
    }
}


calc_errors
calc_compute_bruno(struct collected_data *d, float_matrix **distance_estimates, float_matrix **error_estimates, float_matrix **substitution_rate_matrix, FILE *informationfile)
{
    int verbose;
    unsigned long total, i;
    float_matrix *sum;
    float_matrix *eigenvalues, *eigenvectors; 
    matrix_list *m;
    matrix_errors m_err;
    calc_errors c_err;
    int steady_state_vector;	/* Index of the column in the matrix eigenvectors that represent the limit distribution, the steady state. */
    

    if (informationfile != NULL)
	verbose = 1;
    else 
	verbose = 0;

    /*
      Calculate nucleotide frequencies.
     */
    total = 0;
    for (i = 0; i < end_of_nucl_codes; i++) {
	total += matrix_val(d->base_count, i, i);
    }
    for (i = 0; i < end_of_nucl_codes; i++) {
	matrix_set(d->base_count, i, i, matrix_val(d->base_count, i, i) / total);
    }
    
    /* 
	   Sum all substitution matrices into one matrix.
	   Hopefully, the noise in the matrices will cancel somewhat.
	   The matrices in the freqs list is kept as matrices counting
	   the substitutions, but we are adding copied matrices that 
	   have been transformed into substitution probability matrices
	   P(t) for some time t.
     */
    sum = matrix_init_square(end_of_nucl_codes);

    if (!sum) {
	return C_OUT_OF_MEMORY;
    }

    for (m = d->data; m != NULL; m = m->next) {
	m_err = matrix_add_to(m->probability_estimate, sum);
	if (m_err != M_NO_ERROR) {
	    matrix_free(sum);
	    matrix_free(d->base_count);
	    return C_COMPUTATIONAL_ERROR;
	}
    }
#ifdef DONTWANTTHISNOMORE
    if (verbose) {
	fprintf(informationfile, "\nThe sum of probability matrix estimates:\n");
	matrix_fprint(informationfile, sum);
    }
#endif
    /*
      We want to change basis for a while when enforcing reversibility.
     */
    
    symmetrize(sum, d->base_count);
#ifdef DONTWANTTHISNOMORE
    if (verbose) {
	fprintf(informationfile, "\nThe symmetrized P matrix:\n");
	matrix_fprint(informationfile, sum);
    }
#endif
    /* 
       sum is probably not 'true' symmetric at this time, so we enforce it.
     */
    c_err = average(sum); 
    if (c_err) {
	return c_err;
    }
#ifdef DONTWANTTHISNOMORE
    if (verbose) {
	fprintf(informationfile, "\nThe avereged symmetrized P matrix:\n");
	matrix_fprint(informationfile, sum);
    }
#endif    
    /* 
       Calculate eigenvectors and eigenvalues of the sum matrix (P).
     */

    eigenvectors = matrix_init_square(matrix_size(sum));
    eigenvalues = matrix_init(matrix_size(sum), 1);

    if (!eigenvectors || !eigenvalues) {
	matrix_free(sum);
	matrix_free(d->base_count);
	matrix_free(eigenvectors);
	return C_OUT_OF_MEMORY;
    } else {
	/*
	  Here is the actual eigen finding function called.
	 */
	m_err = matrix_eigen(sum, eigenvectors, eigenvalues);
	if (m_err != M_NO_ERROR) { 
	    return C_COMPUTATIONAL_ERROR;
	} else {
	    /* 
	       Find the eigenvector representing the steady state. 
	     */
	    steady_state_vector = find_steady_state_vector(eigenvectors);
	    if (steady_state_vector < 0) {
		return C_COULD_NOT_FIND_STEADY_STATE;
	    }
	    /*
		  Make sure that the eigenvectors are normalized. I don't trust
		  the eigenvector routine unfortunately.
	     */
	    for (i = 0; i < end_of_nucl_codes; i++) {
		matrix_normalize_column(eigenvectors, i);
	    }

	    d->eigenvectors = eigenvectors;
	    d->steady_state_vector = steady_state_vector;

	    /* 
	       This stuff might actually be nice to have printed somewhere
	       but I leave it out because it is a drag to implement
	       it in a nice way for both a command line program and a GUI.
	     */
	    if (verbose) {
		fprintf(informationfile, "\nEigenvectors in the columns of this matrix:\n");
		matrix_fprint(informationfile, eigenvectors);

		fprintf(informationfile, "\nThe corresponding eigenvalues:\n");
		matrix_fprint(informationfile, eigenvalues);
	
		fprintf(informationfile, "Equilibrium: ");
		for (i=0; i<4; i++) {
		    fprintf(informationfile, "%f ", matrix_val(d->base_count, i, i));
		}
		fprintf(informationfile, "\n");
	    }

	    *distance_estimates = matrix_init_square(d->nr_of_entries);
	    if (!distance_estimates) {
		print_calc_error(C_OUT_OF_MEMORY);
		exit(1);
	    }
	    *error_estimates = matrix_init_square(d->nr_of_entries);
	    if (!error_estimates) {
		print_calc_error(C_OUT_OF_MEMORY);
		exit(1);
	    }

	    *substitution_rate_matrix = matrix_init_square(4);
	    if (!substitution_rate_matrix) {
		print_calc_error(C_OUT_OF_MEMORY);
		exit(1);
	    }

	    c_err = estimate_rates_and_distances(d->data, 
						 d->base_count, 
						 eigenvectors, 
						 steady_state_vector,
						 *distance_estimates, 
						 *error_estimates, 
						 *substitution_rate_matrix);
	    
	    return c_err;
	}
    }
} /* calc_compute_bruno */

calc_errors
calc_compute_felsenstein(struct collected_data *d, float_matrix **distance_estimates, float_matrix **error_estimates, float_matrix **rates, FILE *informationfile)
{
    return C_PROGRAMMING_ERROR;
}
/*
  calc_pairwise_freqs
  For every pair of sequences, create a matrix containing a count of 
  substitutions. A list of the matrices is returned. In the supplied 
  matrix nucl_count, a total count of the different nucleic acids is
  kept. 

  Note that in previous versions, the matrices in the list were scaled
  so that every column summed to one.
*/

matrix_list *
calc_pairwise_freqs(seqlist *entries, float_matrix *nucl_count) {
    seqlist *e1, *e2;
    float_matrix *counts, *probs;
    matrix_list *results=NULL, *tmp, *last=NULL;
    int i;
    nucleotide_code m;

    results = NULL;

    for (e1 = entries; e1 != NULL; e1 = gb_next_entry(e1)) {
	for (i = 0; i < gb_seq_length(e1); i++) {
	    m = nucleotide(gb_seq_site(e1, i));
	    if (m != nucl_not_implemented) {
		matrix_set(nucl_count, m, m, matrix_val(nucl_count, m, m) +1);
	    }
	}      
	for (e2 = gb_next_entry(e1); e2 != NULL; e2 = gb_next_entry(e2)) {
	    counts = matrix_init_square(end_of_nucl_codes);
	    probs = matrix_init_square(end_of_nucl_codes);
	    if (counts && probs) {
		if (differences(e1, e2, counts) == C_NO_ERROR) {
		    if (tmp = (matrix_list*) malloc(sizeof(matrix_list))) {
			tmp->seq1 = gb_locus(e1);
			tmp->seq2 = gb_locus(e2);
			tmp->change_count = counts;
			matrix_copy(counts, probs);
			make_rate_matrix(probs, tmp->base_count);
			tmp->probability_estimate = probs;
			tmp->next = NULL;
			if (last) {
			    last->next = tmp;
			} else {
			    results = tmp;
			}
			last = tmp;
		    } else {
			fprintf(stderr, "Out of memory.\n");
			exit(1);
		    }
		} else {
		    fprintf(stderr, "Error when comparing sequences '%s' and '%s'.\n",
			    gb_locus(e1), gb_locus(e2));
		    exit(1);
		}
	    } else {
		fprintf(stderr, "Null sequence found. Aborting.\n");
		exit(3);
	    }
	}
    }
    return results;
}


matrix_list *
calc_specified_freqs(seqlist *entries, seqnamelist *from, seqnamelist *to
		     , float_matrix *nucl_count) {
    seqlist *e1, *e2;
    seqnameitem *from_name, *to_name;
    float_matrix *matrix;
    matrix_list *results=NULL, *tmp, *last=NULL;
    int i;
    nucleotide_code m;

    results = NULL;
  
    for (from_name = first_seq_name(from); from_name != NULL; from_name = from_name->next) {
	e1 = gb_find_locus(entries, from_name->seqname);
	if (e1) {
	    for (to_name = first_seq_name(to); to_name != NULL; to_name = to_name->next) {
		e2 =  gb_find_locus(entries, to_name->seqname);
		if (e2) {
		    if (matrix = matrix_init_square(end_of_nucl_codes)) {
			if (!differences(e1, e2, matrix)) {
/*	      make_rate_matrix(matrix); There is a new 2-parameter version of this function!!!*/
			    if (tmp = (matrix_list*) malloc(sizeof(matrix_list))) {
				tmp->change_count = matrix;
				tmp->next = NULL;
				if (last) {
				    last->next = tmp;
				} else {
				    results = tmp;
				}
				last = tmp;
			    } else {
				fprintf(stderr, "Out of memory.\n");
				exit(1);
			    }
			} else {
			    fprintf(stderr, "Error when comparing sequences '%s' and '%s'.\n"
				    , gb_locus(e1), gb_locus(e2));
			    exit(1);
			}
		    }
		} else {
		    fprintf(stderr, "Warning: Couldn't find entry '%s' given in direction file.\n", to_name->seqname);
		}
	    }
	} else {
	    fprintf(stderr, "Warning: Couldn't find entry '%s' given in direction file.\n", from_name->seqname);
	}
    }

				/* Count nucleotides */
    for (from_name = first_seq_name(from); from_name != NULL; from_name = from_name->next) {
	e1 = gb_find_locus(entries, from_name->seqname);
	if (e1) {
	    for (i=0; i < gb_seq_length(e1); i++) {
		m = nucleotide(gb_seq_site(e1, i));
		if (m != nucl_not_implemented) {
		    matrix_set(nucl_count, m, m, matrix_val(nucl_count, m, m) +1);
		}
	    }
	}
    }
    for (to_name = first_seq_name(to); to_name != NULL; to_name = to_name->next)
	{
	    e2 = gb_find_locus(entries, to_name->seqname);
	    if (e2) {
		for (i=0; i < gb_seq_length(e2); i++) {
		    m = nucleotide(gb_seq_site(e2, i));
		    if (m != nucl_not_implemented) {
			matrix_set(nucl_count, m, m, matrix_val(nucl_count, m, m) +1);
		    }
		}
	    }
	}

    return results;
}



matrix_list *
calc_paired_freqs(seqlist *entries, seqnamelist *from, seqnamelist *to, seqnamelist *weights, float_matrix *nucl_count) {
    seqlist *e1, *e2;
    seqnameitem *from_name, *to_name;
    float_matrix *matrix;
    matrix_list *results=NULL, *tmp, *last=NULL;
    int i;
    nucleotide_code m;
    seqnamelist *unique_seq_listing;

    results = NULL;

    for (from_name = first_seq_name(from), to_name = first_seq_name(to); 
	 from_name != NULL, to_name != NULL; 
	 from_name = from_name->next, to_name = to_name->next) {
	e1 = gb_find_locus(entries, from_name->seqname);
	e2 = gb_find_locus(entries, to_name->seqname);
	if (e1 && e2) {
	    if (matrix = matrix_init_square(end_of_nucl_codes)) {
		if (differences(e1, e2, matrix) == C_NO_ERROR) {
/*	  make_rate_matrix(matrix); There is a new 2-parameter version of this 
	  function. Also, you probably don't want this call nowadays anyway!!! */
		    if (tmp = (matrix_list*) malloc(sizeof(matrix_list))) {
			tmp->change_count = matrix;
			tmp->next = NULL;
			if (last) {
			    last->next = tmp;
			} else {
			    results = tmp;
			}
			last = tmp;
		    } else {
			fprintf(stderr, "Out of memory.\n");
			exit(1);
		    }
		} else {
		    fprintf(stderr, "Error when comparing sequences '%s' and '%s'.\n", gb_locus(e1), gb_locus(e2));
		    exit(1);
		}
	    }
	} else {
	    fprintf(stderr, "Warning: Couldn't find entry '%s' given in direction file.\n", from_name->seqname);
	}
    }

    /* Count nucleotides */
    /* Put all sequence names in one listing.  */
    unique_seq_listing = (seqnamelist *) malloc (sizeof(seqnamelist));
    if (!unique_seq_listing) {
	fprintf(stderr, "Out of memory! Aborting.\n");
	exit(1);
    }
    init_seq_name_list(unique_seq_listing);
  
    for (from_name = first_seq_name(from); from_name != NULL; from_name = from_name->next) {
	if (item_index(unique_seq_listing, from_name->seqname) < 0) {
	    /* If not already in list, do */
	    append_seq_name(unique_seq_listing, from_name->seqname);
	}
    }

    for (to_name = first_seq_name(to); to_name != NULL; to_name = to_name->next)
	{
	    if (item_index(unique_seq_listing, to_name->seqname) < 0) {
		append_seq_name(unique_seq_listing, to_name->seqname);
	    } 
	}
   
    /* Now the names are all unique */
    for (from_name = first_seq_name(unique_seq_listing); 
	 from_name != NULL; 
	 from_name = from_name->next) {

	e1 = gb_find_locus(entries, from_name->seqname);
	if (e1) {
	    for (i=0; i < gb_seq_length(e1); i++) {
		m = nucleotide(gb_seq_site(e1, i));
		if (m != nucl_not_implemented) {
		    matrix_set(nucl_count, m, m, matrix_val(nucl_count, m, m) +1);
		}
	    }
	}
    }
    delete_list(unique_seq_listing); /* Don't need this one any more */
  
    return results;
}


calc_errors
differences(seqlist *e1, seqlist *e2, float_matrix *changes) {
    int length, i;
    nucleotide_code m, n;
  
    if ((length = gb_seq_length(e1)) != gb_seq_length(e2)) {
	fprintf(stderr, "Warning! Sequences %s and %s have different length.\n", gb_locus(e1), gb_locus(e2));
	length = ((gb_seq_length(e1) < gb_seq_length(e2)) ? gb_seq_length(e1) : gb_seq_length(e2));
    }
     
    if (length > 0) {	/* Something's wrong if 0 or -1 */
	for (i=0; i < length; i++) {
	    m = nucleotide(gb_seq_site(e1,i));
	    n = nucleotide(gb_seq_site(e2,i));
	    if ((m != nucl_not_implemented) && (n != nucl_not_implemented)) {
		matrix_set(changes, m, n, matrix_val(changes, m, n) + 1);
	    }
	}
	return C_NO_ERROR;
    } else {
	return C_BAD_SEQUENCES;			/* Something went wrong */
    }
}

nucleotide_code
nucleotide(char a) {
    switch(tolower(a)) {
    case 'a':
	return nucl_a;
    case 'c':
	return nucl_c;
    case 'g':
	return nucl_g;
    case 't':
	return nucl_t;
    default:
	return nucl_not_implemented;
    }
}



/*
  make_rate_matrix
  
  This function normalizes every column in the given matrix, that is, 
  it divides every column by it's sum, which of course make the column 
  sum to 1. In the second parameter, the array, the number of sites in 
  the first sequence that had base i is counted.

*/


calc_errors
make_rate_matrix(float_matrix *m, unsigned long count[4]) {
    unsigned i, j;

    if (m && count) {
	for (j = 0; j < 4; j++) {
	    count[j] = 0;
	    for (i = 0; i < 4; i++) {
		count[j] += matrix_val(m, i, j);
	    }
	    for (i = 0; i < 4; i++) {
		matrix_set(m, i, j, matrix_val(m, i, j) / count[j]);
	    }
	}
	return C_NO_ERROR;
    } else {
	return C_NULL_MATRIX;
    }
}

	   

/*
  symmetrize
  A substitution rate matrix Q or a substitution probability matrix P is
  expected to be reversibel. The reversibility criterion demands that the 
  product \Pi^{1/2} * Q * \Pi^{-1/2} equals its transpose. This function
  calculates \Pi^{1/2} * M * \Pi^{-1/2} for a matrix M and base frequencies
  \Pi.

  Note that the function is "destructive" and changes M.
*/

calc_errors
symmetrize(float_matrix *M, float_matrix *base_frequencies) {
    int size, i;
    float sqrt_pi;
    float_matrix *left, *right, *tmp;

    if (M) {
	size = matrix_size(M);
	if (size == -1) {
	    return C_WRONG_SIZE_OF_MATRIX;
	}
    } else {
	return C_NULL_MATRIX;
    }

    left = matrix_init_square(size);
    right = matrix_init_square(size);
    tmp = matrix_init_square(size);

    for (i = 0; i<size; i++) {
	sqrt_pi = sqrt(matrix_val(base_frequencies, i, i));
	matrix_set(left, i, i, 1/sqrt_pi);
	matrix_set(right, i, i, sqrt_pi);
    }

    matrix_mul(left, M, tmp);
    matrix_mul(tmp, right, M);
  
    matrix_free(left);
    matrix_free(right);
    matrix_free(tmp);
  
    return C_NO_ERROR;
}


/*
  Make the matrix 'true' symmetric (in contrast to symmetrize()) by
  averaging the elements that should be equal.
  
  The function performs the calculation M := (M + M^t) * 1/2.
 */

calc_errors
average(float_matrix *m) {
    float size, i, j;
    float med;
  
    if (m) {
	size = matrix_size(m);
	if (size == -1) {
	    return C_NOT_A_SQUARE_MATRIX; /* Can only call this function with a squa
					     re matrix */
	} 
    } else {
	return C_NULL_MATRIX;			/* Matrix missing */
    }

    for (i = 0; i < size-1; i++) {
	for (j = i+1; j < size; j++) {
	    med = (matrix_val(m, i, j) + matrix_val(m, j, i)) / 2;
	    matrix_set(m, i, j, med);
	    matrix_set(m, j, i, med);
	}
    }
  
    return C_NO_ERROR;
}

/*
  extract_and_transform_eigens
  This funtion take a matrix containg eigenvectors obtained in our reversibel
  world and extracts them to a 4-array of vectors (e_vectors) and also
  transform them back to the "original world", the world where we obtained 
  the data in the first place.
*/
  
calc_errors
extract_and_transform_eigens(float_matrix *eigenvectors, 
			     float_matrix *left_e_vectors[end_of_nucl_codes], 
			     float_matrix *right_e_vectors[end_of_nucl_codes], 
			     float_matrix *base_freqs) /* Diagonal matrix */
{
    float sqrt_base_freq;
    int size;
    float_matrix *invSqrtPi, *sqrtPi, *vector, *tmp;
    int i;
    matrix_errors m_err;

    if (eigenvectors) {
	size = matrix_size(eigenvectors);
	if (size == -1) {
	    return C_NOT_A_SQUARE_MATRIX;
	}
    } else {
	return C_NULL_MATRIX;
    }
  
    invSqrtPi = matrix_init_square(size);	/* \Pi^{-1/2} */
    sqrtPi = matrix_init_square(size);	/* \Pi^{1/2} */
  
    if (!sqrtPi || !invSqrtPi) {
	return C_OUT_OF_MEMORY;
    }

    for (i = 0; i<size; i++) {
	sqrt_base_freq = sqrt(matrix_val(base_freqs, i, i));
	matrix_set(invSqrtPi, i, i, 1/sqrt_base_freq);
	matrix_set(sqrtPi, i, i, sqrt_base_freq);
    }
  
    vector = matrix_init(4,1);
    if (!vector) {
	return C_OUT_OF_MEMORY;
    }

    for (i = 0; i<size; i++) {
	matrix_copy_column(eigenvectors, i, vector, 0);
	tmp = matrix_make_transpose(vector); /* 1x4 --- tmp = vector^T */
	if (!tmp) {
	    return C_OUT_OF_MEMORY;
	}
	m_err  = matrix_mul(tmp, invSqrtPi, left_e_vectors[i]);
	if (m_err != M_NO_ERROR) {
	    return C_COMPUTATIONAL_ERROR;
	}
	m_err = matrix_mul(sqrtPi, vector, right_e_vectors[i]);
	if (m_err != M_NO_ERROR) {
	    return C_COMPUTATIONAL_ERROR;
	}
	matrix_free(tmp);
    }
    matrix_free(invSqrtPi);
    matrix_free(sqrtPi);
    matrix_free(vector);
  
    return C_NO_ERROR;
} /* extract_and_transform_eigens */


/* 
  This function calculates the variance of the distance measures.
  Your return values are stored in the diagonal of a 3x3 matrix that 
  you supply initialized. This peculiar way of storing the variances is 
  due the fact that we will want to have a covariance matrix later, and
  this functions is supposed to be used in conjunction with 
  'distance_covariance'. The "eigenvectors" are supposed to have been 
  transformed by the \sqrt(\pi) transform. Several of the variable 
  names refer to a notation that Bill used in a note to Lars. 

 */

calc_errors
distance_variance(float distance[3], /* Really "const * time" */
		  float_matrix *left_eigen[4], /* 4 vectors of size 1x4(?) */
		  float_matrix *right_eigen[4], /* 4 vectors of size 4x1(?) */
		  float_matrix *eigenvalues, /* 1x4 vector */
		  float_matrix *substitution_probability, /* 4x4 matrix */
		  unsigned long base_count[4],
		  int steady_state_vector,
		  float_matrix *return_values) { /* 3x3 vector */
    int i, j, k;
    int n;			/* Sets which of the three distances that */
				/* is being investigated. */
    float v_square;
    float isum, jksum;
    float difference, denominator;
    float_matrix *tmp_vector, *tmp_scalar; /* Work space */
    int index;
    
    if (base_count && eigenvalues && substitution_probability &&
	left_eigen && right_eigen && return_values) {
	tmp_vector = matrix_init(4, 1);
	tmp_scalar = matrix_init(1, 1);
	if (!tmp_scalar || !tmp_vector) {
	    return C_OUT_OF_MEMORY;
	}

	/* Start calculating the three different variances */
	index = 0;
	for (n = 0; n < 4; n++) {
	    if (n != steady_state_vector) {
		isum = 0;
		for (i = 0; i < 4; i++) {
		    v_square = matrix_val(right_eigen[n], i, 0);
		    v_square = v_square * v_square;
		    
		    jksum = 0;
		    for (j = 0; j < 3; j++) {
			for (k = j+1; k < 4; k++) {
			    difference = matrix_val(left_eigen[n], 0, j) - matrix_val(left_eigen[n], 0, k);
			    jksum += difference * difference
				* matrix_val(substitution_probability, i, j)
				* matrix_val(substitution_probability, i, k);
			}
		    } /* j-k loop -- doing second sum */
		    isum += v_square * jksum / base_count[i];
		    
		}	/* i loop -- doing first sum */
		
/*      denominator = distance[n-1] * distance[n-1] * matrix_val(eigenvalues, n, 0)  * matrix_val(eigenvalues, n, 0)  * matrix_val(eigenvalues, n, 0) * matrix_val(eigenvalues, n, 0);
	matrix_set(return_values, n-1, n-1, isum / denominator);
	*/
		denominator = exp(-2* distance[index] * matrix_val(eigenvalues, n, 0)) * matrix_val(eigenvalues, n, 0) * matrix_val(eigenvalues, n, 0);
		matrix_set(return_values, index, index, isum / denominator);
		index++;
	    }
	} /* n loop -- doing the three variances */
    
	return C_NO_ERROR;

    } else {
	return C_NULL_MATRIX;
    }  
} /* distance_variance */



/* 
  This function calculates the *covariance* of the distance measures.
  Your return values are stored in the 3x3 matrix that you supply
  initialized. The "eigenvectors" are supposed to have been transformed
  by the \sqrt(\pi) transform. Several of the variable names refer
  to a notation that Bill used in a note to Lars. 

  Note that this function will leave the matrix diagonal untouched. There, 
  the variance of a distance should be found.
*/

calc_errors
distance_covariance(float distance[3], /* Really "const * time" */
		    float_matrix *left_eigen[4], /* 4 vectors of size 1x4 */
		    float_matrix *right_eigen[4], /* 4 vectors of size 4x1 */
		    float_matrix *eigenvalues, /* 1x4 vector */
		    float_matrix *substitution_probability, /* 4x4 matrix */
		    unsigned long base_count[4], 
		    int steady_state_vector,
		    float_matrix *return_values) 
{
    int i, j, k;
    int m, n;			/* Sets which of the three distances that */
				/* is being investigated. */
    float dm_dn;		/* Product of two distance measures */
    float vm_vn;
    float isum, jksum;
    float_matrix *tmp_vector, *tmp_scalar; /* Work space */
    int indexm, indexn;
    
    if (base_count && eigenvalues && substitution_probability &&
	left_eigen && right_eigen && return_values) {
	tmp_vector = matrix_init(4, 1);
	tmp_scalar = matrix_init(1, 1);
	if (!tmp_scalar || !tmp_vector) {
	    return C_OUT_OF_MEMORY;
	}

	/* Start calculating the six different covariances */
	indexm = 0;
	for (m = 0; m < 3; m++) {
	    if (m != steady_state_vector) {
		indexn = indexm+1;
		for (n = m+1; n < 4; n++) {
		    if (n != steady_state_vector) {
			dm_dn = exp(-distance[indexm] * matrix_val(eigenvalues, m, 0))
			    * exp(-distance[indexn] * matrix_val(eigenvalues, n, 0)); 
			
			isum = 0;
			for (i = 0; i < 4; i++) {
			    vm_vn = matrix_val(right_eigen[m], i, 0) * matrix_val(right_eigen[n], i, 0);
			    jksum = 0;
			    for (j = 0; j < 3; j++) {
				for (k = j+1; k < 4; k++) {
				    jksum += (matrix_val(left_eigen[m], 0, j) - matrix_val(left_eigen[m], 0, k))
					* (matrix_val(left_eigen[n], 0, j) - matrix_val(left_eigen[n], 0, k))
					* matrix_val(substitution_probability, i, j)
					* matrix_val(substitution_probability, i, k);
				}
			    } /* j-k loop -- doing second sum */
			    isum += vm_vn * jksum / base_count[i];
			} /* i loop -- doing first sum */
			
			isum = isum / (dm_dn * matrix_val(eigenvalues, m, 0) * 
				       matrix_val(eigenvalues, n, 0));
			matrix_set(return_values, indexm, indexn, isum);
			matrix_set(return_values, indexn, indexm, isum);
			indexn++;
		    }
		} /* n loop */
		indexm++;
	    }
	} /* m loop */

	return C_NO_ERROR;
    } else {
	return C_NULL_MATRIX;
    }
} /* distance_covariance */



/*
  least_square_distance
  Used by 'estimate_rates_and_distances' to compute our final distance
  estimate between to sequences. 

*/
calc_errors
least_square_distance(float distance[3], float_matrix *inv_covariance, float *estimate)
{
    matrix_errors merr;		/* For error control */
    float_matrix *d;
    float_matrix *ones, *ones_transp; /* Supposed to be all ones */
    float constant_ones[] = {1.0, 1.0, 1.0};
    float_matrix *tmpvector, *tmpscalar;	/* Working areas */


    /* Initializations */
    d = matrix_init(3, 1);
    if (!d) {
	return C_OUT_OF_MEMORY;
    }

    merr = matrix_set_all(d, (float*) distance);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    ones = matrix_init(1, 3);
    ones_transp = matrix_init(3, 1);
    if (!ones || !ones_transp) {
	return C_OUT_OF_MEMORY;
    }

    merr = matrix_set_all(ones, constant_ones);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    merr = matrix_set_all(ones_transp, constant_ones);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    tmpvector = matrix_init(1, 3);
    tmpscalar = matrix_init(1, 1);
    if (!tmpvector || !tmpscalar) {
	return C_OUT_OF_MEMORY;
    }

    /* Computations */
    merr = matrix_mul(ones, inv_covariance, tmpvector);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    merr = matrix_mul(tmpvector, d, tmpscalar);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    *estimate = matrix_val(tmpscalar, 0, 0);

    merr = matrix_mul(tmpvector, ones_transp, tmpscalar);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    *estimate = *estimate / matrix_val(tmpscalar, 0, 0);

    return C_NO_ERROR;
} /* least_square_distance */

/*
   sqrt_distance_variance
   In contrast to distance_variance, this function computes an 
   error_estimate for the least squares distance for two 
   sequences, and not for any of the distances used when computing
   the final estimate.
*/
   
calc_errors
sqrt_distance_variance(float_matrix *inv_covariance, float *error_estimate) {
    float_matrix *one_vector, *tmp_scalar, *tmp_vector;
    float_matrix *one_vector_transp;
    float ones[] = {1.0, 1.0, 1.0};
    matrix_errors merr;

    tmp_scalar = matrix_init(1, 1);
    one_vector = matrix_init(1, 3);
    one_vector_transp = matrix_init(3, 1);
    tmp_vector = matrix_init(1, 3);
    if (!tmp_scalar || !one_vector || !one_vector_transp || !tmp_vector) {
	return C_OUT_OF_MEMORY;
    }

    merr = matrix_set_all(one_vector, ones);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    merr = matrix_set_all(one_vector_transp, ones);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    merr = matrix_mul(one_vector, inv_covariance, tmp_vector);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    merr = matrix_mul(tmp_vector, one_vector_transp, tmp_scalar);
    if (merr != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
  
    *error_estimate = sqrt(1.0 / matrix_val(tmp_scalar, 0, 0)); 
    return C_NO_ERROR;
} /* sqrt_distance_variance */


/*
  find_substitution_rate_matrix
  This function takes a matrix with elements 0, 1, l3/l2, l4/l2 
  (or similar) on its diagonal and restores it to a substitution 
  rate matrix using the spectral resolution equation.
  
  This function is more or less a mess right now. I will clean it up
  some day when I feel happy.
*/

calc_errors
find_substitution_rate_matrix(float_matrix *eigenvectors,
			      float_matrix *equilibrium_matrix,
			      int steady_state_vector, 
			      float_matrix *estimate)
{
    int i;
    matrix_errors m_err;		/* For error control */
    float sum;			/* the substitution rate matrix. */
    float_matrix *equilibrium_vector;
    float_matrix *equilibrium_sqrt_vector;
    float_matrix *equilibrium_sqrt_matrix;
    float_matrix *equilibrium_inv_sqrt_matrix;
    float_matrix *transformation, *transformation_transposed;
    float_matrix *tmp1, *tmp2, *tmp_vector;
  
    equilibrium_vector = matrix_init(4, 1);
    equilibrium_sqrt_matrix = matrix_init_square(4);
    equilibrium_inv_sqrt_matrix = matrix_init_square(4);
    equilibrium_sqrt_vector = matrix_init(4,1);

    if (!equilibrium_sqrt_vector || !equilibrium_inv_sqrt_matrix
	|| !equilibrium_sqrt_matrix || !equilibrium_vector) {
	return C_OUT_OF_MEMORY;
    }

    for (i = 0; i < 4; i++) {
	matrix_set(equilibrium_vector, i, 0, 
		   matrix_val(equilibrium_matrix, i, i));
	matrix_set(equilibrium_sqrt_vector, i, 0, 
		   sqrt(matrix_val(equilibrium_matrix, i, i)));
	matrix_set(equilibrium_sqrt_matrix, i, i, 
		   matrix_val(equilibrium_sqrt_vector, i, 0));
	matrix_set(equilibrium_inv_sqrt_matrix, i, i, 
		   1/matrix_val(equilibrium_sqrt_vector, i, 0));
    }

#ifdef DEBUG
    printf("\nThe starting values in the rate estimate matrix:\n");
    matrix_print(estimate);
#endif DEBUG

#ifdef DEBUG
    printf("\nThe roots of the equilibrium probs on the diagonal:\n");
    matrix_print(equilibrium_sqrt_matrix);
    printf("\nThe inverse roots of the equilibrium probs on the diagonal:\n");
    matrix_print(equilibrium_inv_sqrt_matrix);
#endif DEBUG

    transformation = matrix_init_square(4);
    transformation_transposed = matrix_init_square(4);
  
    m_err = matrix_copy(eigenvectors, transformation); /* For left multiplication */
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    m_err = matrix_copy_column(equilibrium_sqrt_vector, 0, transformation, steady_state_vector);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

    m_err = matrix_copy(transformation, transformation_transposed);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
    m_err = matrix_transpose(transformation); 
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }

#ifdef DEBUG
    printf("\nThe left and right transformation matrices, containing eigenvectors and base freqs.\nRight:\n");
    matrix_print(transformation);
    printf("\nLeft:\n");
    matrix_print(transformation_transposed);
#endif DEBUG

    tmp1 = matrix_init_square(4);
    tmp2 = matrix_init_square(4);

    m_err = matrix_mul(transformation_transposed, estimate, tmp1);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
#ifdef DEBUG
    printf("\nAfter first left multiplication:\n");
    matrix_print(tmp1);
#endif DEBUG

    m_err = matrix_mul(tmp1, transformation, tmp2);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
#ifdef DEBUG
    printf("\nAfter first right multiplication:\n");
    matrix_print(tmp2);
#endif DEBUG

    m_err = matrix_mul(tmp2, equilibrium_inv_sqrt_matrix, tmp1);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
#ifdef DEBUG
    printf("\nAfter second right multiplication (de-symmetrizing):\n");
    matrix_print(tmp1);
#endif DEBUG

    m_err = matrix_mul(equilibrium_sqrt_matrix, tmp1, estimate);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
#ifdef DEBUG
    printf("\nAfter second left multiplication (de-symmetrizing):\n");
    matrix_print(estimate);
    printf("\n");
#endif DEBUG

#ifdef DEBUG 
    printf("\nThe rate matrix out off the shelf:\n");
    m_err = matrix_print(estimate);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
#endif
  

    /*
      The matrix 'estimate' now contains a reasonable estimate of the
      substitution rate matrix -- up to a factor. For various reasons,
      e.g. for aesthetic reasons we probably want to multiply with a 
      factor of -1 and because we may want to compare the matrix with 
      others estimate, we choose to multiply with a factor that makes 
      the average total substitution frequency equal 1. This is done 
      both by Felsenstein and Yang. 

      tmp1 will contain a copy of our estimate, but we will set the
      diagonal to zero, so that we can make an easy average by 
      multiplying with our estimate of the base frequencies. When we 
      have done that average, we simply multiply our estimate with
      the inverse average.
      */

  
    m_err = matrix_copy(estimate, tmp1);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
    for (i=0; i<4; i++) {
	matrix_set(tmp1, i, i, 0);
    }

    tmp_vector = matrix_init(4, 1);
    if (!tmp_vector) {
	return C_OUT_OF_MEMORY;
    }
    m_err = matrix_mul(tmp1, equilibrium_vector, tmp_vector);
    if (m_err != M_NO_ERROR) {
	return C_COMPUTATIONAL_ERROR;
    }
  
    sum = 0;
    for (i=0; i<4; i++) {
	sum += matrix_val(tmp_vector, i, 0);
    }

    matrix_scalar_mul(estimate, 1/sum);

    return C_NO_ERROR;
}


/* 
   This function estimates a substitution rate matrix and
   pairwise distances and their errors.

   m contains the list of substitution probabilities for the 
   sequence pairs. equilibrium_matrix is  a diagonal matrix
   with equilibrium  frequencies. eigenvectors is a matrix 
   containing the eigenvectors as columns. 
 */

calc_errors
estimate_rates_and_distances(matrix_list *substitution_prob_list,
			     float_matrix *equilibrium_matrix, /* Estimated nucleotide frequencies, 4x4 */
			     float_matrix *eigenvectors,	/* A 4x4 matrix with the vectors as columns */
			     int steady_state_vector, /* the eigenvector representing steady state */
			     float_matrix *distance_matrix,/* For returnvalues */
			     float_matrix *error_matrix,	/* For returnvalues */
			     float_matrix *rate_matrix_estimate) /* For returnvalues */
{
    matrix_errors m_err;		/* For error control */
    calc_errors c_err;		/* For error control */
    int nr_of_sequences;		/* To know when to start a new row in result matrix. */
    unsigned seq1, seq2;		/* Knows which pair of sequences we are working with. */
    matrix_list *substitution_probabilities; 
				/* To iterate through sequence pair data with. */
    float_matrix *current_matrix;	/* A copy of current substitution matrix. */
    float_matrix *eigenvalues;	/* Eigenvalues to the symmetric rate matrix. */
    float_matrix *right_eigenvectors[4];
				/* Four 4x1 vectors. */
    float_matrix *left_eigenvectors[4];
				/* Four 1x4 vectors. */
    float e_to_eigen_times_time;	/* e^{v_i * t} */
    float distance[3];		/* Keep track of current seq pairs three distances. */
    float distance_estimate;	/* Least squares value of the three distances above. */
    float error_estimate;	/* The square root of the variance of the distance. */
    float_matrix *covariance;	/* 3x3 matrix for variance and covariance. */
    float_matrix *inv_covariance; /* Dito, for its inverse. */
    float sumxy[3][3], sumxx[3]; 
				/* These are for the least squares estimate of rate matrix. */
    int i, j;			/* Index variables in loops */
    float_matrix *tmp_scalar, *tmp_vector;
    float tmp[3];

    /* 
       First, check the input somewhat. 
       */
    if (!substitution_prob_list || !equilibrium_matrix || !right_eigenvectors 
	|| !left_eigenvectors || !distance_matrix || !error_matrix) {
	return C_NULL_MATRIX;
    }
  
    nr_of_sequences = matrix_size(distance_matrix);
    if (nr_of_sequences != matrix_size(error_matrix) || nr_of_sequences == -1) {
	return C_WRONG_SIZE_OF_MATRIX; /* Perhaps not the best of error messages, */
				/* but it shouldn't happen... */
    }
  
    /*
      Second, do some preparation with eigenvalues and stuff. 
     
      Extract eigenvectors and multiply them with sqrt(\Pi)
      matrices, i.e. transform back to the original basis.
      We then distingiush between left and right eigenvectors.
      */
 
    for (i=0; i<4; i++) {
	left_eigenvectors[i] = matrix_init(1, 4);
	right_eigenvectors[i] = matrix_init(4, 1);
	if (!left_eigenvectors[i] || !right_eigenvectors[i]) {
	    print_calc_error(C_OUT_OF_MEMORY);
	    exit(1);
	}
    }
    c_err = extract_and_transform_eigens(eigenvectors, 
					 left_eigenvectors, 
					 right_eigenvectors, 
					 equilibrium_matrix);
    if (c_err != C_NO_ERROR) {
	print_calc_error(c_err);
	exit(1);
    }

    /*
      This double loop initializes variables used when estimating
      the substitution rate matrix.
      */
    for (i = 0; i < 3; i++) {
	sumxx[i] = 0;
	for (j = 0; j < 3; j++) {
	    sumxy[i][j] = 0;
	}
    }

/* 
   Create workspace for computing distances. We need a temporary 
   vector and a temporary place for a scalar, resulting from a 
   inner product.
   */
    tmp_scalar = matrix_init(1, 1);
    tmp_vector = matrix_init(1, 4);
    current_matrix = matrix_init_square(4);
    if (!tmp_scalar || !tmp_vector || !current_matrix) {
	return C_OUT_OF_MEMORY;
    }
  
/*
  Go through every sequence pair and calculate distance estimates,
  variance, covariance and prepair for estimating substitution
  rate matrix.
  */
    for (substitution_probabilities = substitution_prob_list; 
	 substitution_probabilities != NULL; 
	 substitution_probabilities = substitution_probabilities->next) {
	j=0;
	for (i=0; i<4; i++) {
	    if (i != steady_state_vector) {
		/* Here goes v_i*P */
		m_err = matrix_mul(left_eigenvectors[i], substitution_probabilities->probability_estimate, tmp_vector);
		if (m_err != M_NO_ERROR) {
		    print_matrix_error(m_err);
		    exit(1);
		}
		/* And here comes the right side: (v_i*P)*v_i^+ */
		m_err = matrix_mul(tmp_vector, right_eigenvectors[i], tmp_scalar);
		if (m_err != M_NO_ERROR) {
		    print_matrix_error(m_err);
		    exit(1);
		}
		e_to_eigen_times_time = matrix_val(tmp_scalar, 0, 0);
		if (e_to_eigen_times_time <= 0) {
		    fprintf(stderr, "Oops, a bad distance was found. Outputting distance -1 there.\n");
		    distance[j] = -1;
		} else {
		    distance[j] = - (float) log(e_to_eigen_times_time); /* = eigenval_i * time */
		}
#ifdef DEBUG_RATIO
		printf("%f\t", distance[j]);
#endif
		j++;
	    } /* if not steady state vector */
	}	/* i loop */
#ifdef DEBUG_RATIO
	printf("\n");
#endif

	/*
	  Here is some work for the substitution rate matrix estimate. We are
	  preparing to come up with a least squares approximation of the 
	  ratios between eigenvalues.
	  */
	for (i = 0; i < 3; i++) {
	    sumxx[i] += distance[i]*distance[i];
	    for (j = i+1; j < 3; j++) {
		sumxy[i][j] += distance[j]*distance[i];
	    }
	}	/* ij loop */
    }  /* All sequence pairs done */

    /*
      It is time to estimate the substitution rate matrix.
      */

#ifdef DEBUG
    printf("Here is the eigenvalue quotients:\n");
    printf("l3/l2 = %f, l4/l2 = %f, l4/l3 = %f\n", 
	   sumxy[0][1]/sumxx[0],
	   sumxy[0][2]/sumxx[0],
	   sumxy[1][2]/sumxx[1]);
#endif DEBUG

    tmp[0] = 1;
    tmp[1] = sumxy[0][1]/sumxx[0];
    tmp[2] = sumxy[0][2]/sumxx[0];

    eigenvalues = matrix_init(4, 1);
    if (!eigenvalues) {
	return C_OUT_OF_MEMORY;
    }

    j=0; 
    for (i=0; i<4; i++) { 
	if (i == steady_state_vector) { 
	    matrix_set(rate_matrix_estimate, i, i, 0); 
	    matrix_set(eigenvalues, i, 0, 0);  
	} else { 
	    matrix_set(rate_matrix_estimate, i, i, tmp[j]); 
	    matrix_set(eigenvalues, i, 0, tmp[j]); 
	    j++; 
	} 
    } 

    /*
      In the following call, 'rate_matrix_estimate' will be adjusted so 
      that it contains the estimated substitution rate matrix.
      */
    c_err = find_substitution_rate_matrix(eigenvectors, equilibrium_matrix, steady_state_vector, rate_matrix_estimate);
    if (c_err != C_NO_ERROR) {
	return c_err;
    }

    /*
      It is time to compute variance, covariance and the least squares
      distance estimate for the two sequences we are currently looking at.
      We have to go through the sequences again. This time, we have a
      good estimate of the eigenvalues, which we will use. 
    
      Variance
      */

    covariance= matrix_init(3, 3);
    if (covariance == NULL) {
	fprintf(stderr, "Out of memory.\n");
	exit(1);
    }
    inv_covariance= matrix_init(3, 3);
    if (inv_covariance == NULL) {
	fprintf(stderr, "Out of memory.\n");
	exit(1);
    }
    seq1 = 0;
    seq2 = 0;
    for (substitution_probabilities = substitution_prob_list; 
	 substitution_probabilities != NULL; 
	 substitution_probabilities = substitution_probabilities->next) {
	seq2++;
	if (seq2 == nr_of_sequences) {
	    seq1++;
	    seq2 = seq1 + 1;
	}

	j=0;
	for (i=0; i<4; i++) {
	    if (i != steady_state_vector) {
		/* Here goes v_i*P */
		m_err = matrix_mul(left_eigenvectors[i], substitution_probabilities->probability_estimate, tmp_vector);
		if (m_err != M_NO_ERROR) {
		    print_matrix_error(m_err);
		    exit(1);
		}
		/* And here comes the right side: (v_i*P)*v_i^+ */
		m_err = matrix_mul(tmp_vector, right_eigenvectors[i], tmp_scalar);
		if (m_err != M_NO_ERROR) {
		    print_matrix_error(m_err);
		    exit(1);
		}
		e_to_eigen_times_time = matrix_val(tmp_scalar, 0, 0);
		if (e_to_eigen_times_time <= 0) {
		    fprintf(stderr, "Oops, a bad distance was found. Outputting distance -1 there.\n");
		    distance[j] = -1;
		} else {
		    distance[j] = - log(e_to_eigen_times_time) / matrix_val(eigenvalues, i, 0);
		}
#ifdef DEBUG
		printf("%f\t", distance[j]);
#endif
		j++;
	    } /* if not steady state vector */
	}	/* i loop */
#ifdef DEBUG
	printf("\n");
#endif

	c_err = distance_variance(distance, 
				  left_eigenvectors, 
				  right_eigenvectors,
				  eigenvalues, 
				  substitution_probabilities->probability_estimate,
				  substitution_probabilities->base_count, 
				  steady_state_vector, 
				  covariance);
	if (c_err != C_NO_ERROR) {
	    return c_err;
	}

	/*
	  Co-variance plus its inverse which is used in computing the 
	  distance estimate.
	  */
	c_err = distance_covariance(distance, 
				    left_eigenvectors, 
				    right_eigenvectors,
				    eigenvalues, 
				    substitution_probabilities->probability_estimate,
				    substitution_probabilities->base_count, 
				    steady_state_vector, covariance);
	if (c_err != C_NO_ERROR) {
	    return c_err;
	}

	c_err = matrix_inverse(covariance, inv_covariance);
	if (c_err != C_NO_ERROR) {
	    return c_err;
	}
	/*
	  Distance estimate. Store it in the result matrix, of course.
	  */
	c_err = least_square_distance(distance, inv_covariance, &distance_estimate);
	if (c_err != C_NO_ERROR) {
	    return c_err;
	}
	m_err = matrix_set(distance_matrix, seq1, seq2, distance_estimate);
	if (m_err != M_NO_ERROR) {
	    return C_COMPUTATIONAL_ERROR;
	}
	m_err = matrix_set(distance_matrix, seq2, seq1, distance_estimate);
	if (m_err != M_NO_ERROR) {
	    return C_COMPUTATIONAL_ERROR;
	}
    
	/*
	  Calculate error estimate.
	  */
	c_err = sqrt_distance_variance(inv_covariance, &error_estimate);
	if (c_err != C_NO_ERROR) {
	    return c_err;
	}
    
	m_err = matrix_set(error_matrix, seq1, seq2, error_estimate);
	if (m_err != M_NO_ERROR) {
	    return C_COMPUTATIONAL_ERROR;
	}
	m_err = matrix_set(error_matrix, seq2, seq1, error_estimate);
	if (m_err != M_NO_ERROR) {
	    return C_COMPUTATIONAL_ERROR;
	}
    }	/* All sequence pairs done again! */

    return C_NO_ERROR;

} /* estimate_rates_and_distances */


/*
  find_steady_state_vector
  I discovered that the index of the eigenvector representing the
  steady state is not always the same... I think it depends on 
  computer architecture for some reason.
  
  This function finds the steady state vector by looking for an
  all positive eigen vector. It returns the index of the first such
  vector (column) it finds in the given matrix.

  If no such column is found, -1 is returned.
 */
int 
find_steady_state_vector(eigenvectors)
    float_matrix *eigenvectors;
{
    int row, column;
    int status;

    for (column=0; column<4; column++) {
	status = 1;
	for (row=0; row<4; row++) {
	    if (matrix_val(eigenvectors, row, column) < 0) {
		status = 0;
	    }
	}
	if (status == 1) {
	    return column;
	}
    }
    return -1;
}



char *calc_error_codes[] = {
    "No error.",
    "Out of memory.",
    "Not a square matrix.",
    "Wrong size of matrix.",
    "Null matrix encountered.",
    "Bad sequences found. Check your input.",
    "Computational error. Bad programming probably!",
    "Could not find eigenvector representing the steady state.",
    "Error."
};				/* calc_error_codes */

void
print_calc_error(calc_errors errflag) {
    if (errflag < C_END_OF_ERRORS) {
	fprintf(stderr, "\n%s\n", calc_error_codes[errflag]);
    } else {
	fprintf(stderr, "\nUnknown error. Don't panic.\n");
    }
}

char * 
calc_error_string(calc_errors error_code) {
    if (error_code < C_END_OF_ERRORS) {
	return calc_error_codes[error_code];
    } else {
	return calc_error_codes[C_END_OF_ERRORS];
    }
}
