/*
Copyright, 1994-1995, The Regents of the University of California. 
This software was produced under U.S. Government contract(XXXXX)
by Los Alamos National Laboratory, which is operated by the University 
of California for the U.S. Department of Energy. The U.S Government 
is licensed to use, reproduce, and distribute this software. Permission 
is granted to the public to copy and use this software without charge, 
provided that this Notice and any statement of authorship are reproduced 
on all copies. Neither the Government nor the University makes any 
warranty, express or implied, or assumes any liability or responsibility 
for the use of this software.
*/

/* $Id: main.c,v 1.4 96/09/11 15:17:30 arve Exp $ */
#define REVISION "$Revision: 1.4 $"

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
//++pac: strdup()
#ifdef _PLAN9
#include <bsd.h>
#endif

#include "version.h"
#include "matrix.h"
#include "structures.h"
#include "struct_methods.h"

#ifdef DEBUG_STRIP
#include <memory.h>
#endif DEBUG_STRIP


#include "seqio.h"
#include "seqlist.h"
#include "calc.h"

#define DEFAULT_ALLOWED "acgtACGT"
//++pac
#ifdef _PLAN9
static char lanl_copyright[] = "see copyright-notice.txt"
;
static char seqio_copyright[] =   "see seqio_copyright_notice.txt"
;
static char cephes_copyright[] = "Cephes eigen function: Copyright 1973, 1991 by Stephen L. Moshier; Copyleft version.";
#else
static char lanl_copyright[] = 
#include "copyright-notice.txt"
;
static char seqio_copyright[] = 
#include "seqio_copyright_notice.txt"
;
static char cephes_copyright[] = "Cephes eigen function: Copyright 1973, 1991 by Stephen L. Moshier; Copyleft version.";
#endif



enum return_status {
  R_OK = 0x0,
  R_DIFFERENT_LENGTHS = 0x1,
  R_OUT_OF_MEMORY = 0x2,
  R_NO_DATA = 0x4,
  R_FOUND_AMBIGOUITY = 0x8,
  R_END_OF_CODES = 0xFF
};


enum return_status
strip_indels_etc(struct collected_data d, char *allowed, seqlist **new_list)
{
    int status = R_OK;
    seqlist *e, *entries;
    int length;
    int  found_allowed;
    int bar, removed_sites, i;
    char current;

    if (d.entries == NULL) {
      return R_NO_DATA;
    }

    entries = sl_copy_list(d.entries);
    e = entries;
    if (e == NULL) {
      status = R_OUT_OF_MEMORY;
      return status;
    }

    length = strlen(e->seq);
    while (e->next) {
      if (strlen(e->seq) != strlen(e->next->seq)) {
	status |= R_DIFFERENT_LENGTHS;
      }
      if (length > strlen(e->seq)) {
	length = strlen(e->seq);
      }
      e = e->next;
    } 

/*     if (warning) { */
/*       fprintf(stderr, "Sequences have different length! Stripping down to the shortest sequence.\n"); */
/*     } */

/*     if (verbose) { */
/*       fprintf(stderr, "Sequences are %ld bp long.\n", length); */
/*     } */

    bar = 0;
    /* 
       This bar will be moved upwards every time a good cross section is found
       */
    removed_sites = 0;
    /* 
       Remove site if any entry contains a forbidden character.
       Study cross section by cross section. If we find a bad cross
       section, go to next a copy next good cross section over the
       first 'available' or 'not used' cross section.
       */
    for (i = 0; i < length; i++) {
#ifdef DEBUG
      fprintf(stderr, "%d\n", i);
#endif DEBUG
      status = NULL; 
      found_allowed = 1;
      for (e = entries;	/* , current = e->seq[i] */
	   (e != NULL) && (current = e->seq[i]);
	   e = e->next) {
	if (strchr(allowed, current) == NULL) {/* Found forbidden */
	  found_allowed = 0;
	  break;		/* Escape inner loop */
	} else {
	  if (bar < i) {	/* Only copy if necessary */
	    e->seq[bar] = current;   
	  }
	}
      }
      if (found_allowed) {	/* If we found a good cross section, */
	bar++;		/* move up the bar one step! */
      } else {
	removed_sites++;
      }
    }
    
    for (e = entries; e != NULL; e = e->next) {
      e->seq[length - removed_sites] = '\0';
    }
    
    *new_list = entries;
    return status;
}

void
main(int argc, char **argv) {
    int verbose; /* Flags for cmd line args */
    int keep_seqio_prefix, shorten_long_seqnames, print_copyright_notice;
    int errflag=0, c, i, input_given=0;
    calc_errors c_err;
    time_t current_time;
    FILE *outfile, *informationfile;
    char format[80];		/* Asumed max size of seq. identifier. */
    int name_length;
    char *outfilename; /* Only used if verbose==1 */
    char *informationfilename;	/* Not a very good name, 'information...' */
    seqlist *current_entry;
    float_matrix *distance_estimates, *error_estimates, *substitution_rate_matrix;
    struct collected_data data;
    char *allowed_characters, *rev_string;
    enum return_status r_val;

    SEQFILE *sfp;
    seqlist* new_entries;
#ifdef DEBUG_STRIP
    SEQINFO info;
    seqlist *e;
#endif DEBUG_STRIP

    char *optarg;		/* For commandline option parsing */
    int option;			/* For commandline option parsing */

    /* Default values */

    verbose = 0;
    data.entries = NULL; 
    outfilename = NULL;
    informationfilename = NULL;
    informationfile = stdout;
    allowed_characters = DEFAULT_ALLOWED;
    keep_seqio_prefix = 0;
    shorten_long_seqnames = 0;
    print_copyright_notice = 0;

    /*
     * Handle commandline options.
     */

    option = 0;
    while (option < argc - 1) {
	option++;
	if (*argv[option] == '-') {
	    c = *(argv[option]+1);
	    if (option < argc - 1) {
		optarg = argv[option+1]; 
	    } else {
		optarg = "";	/* Empty string */
	    }
	} else {		/* unexpected argument */
 	    c = '\0';		/* Assume input file */
	    optarg = argv[option]; 
	}
 
	switch (c) {
	case 'c':
	    print_copyright_notice = 1;
	    break;

	case 'x':
	    if (!(informationfile = freopen(optarg, "w", stderr))) {
		fprintf(stderr, "Couldn't open '%s' for writing.\n", optarg);
		exit(1);
	    }
	    informationfilename = strdup(optarg);
	    option++;
	    break;
 
	case '-':
	    input_given = 1;
	    sfp = seqfopen("-", "r", NULL); /* No format assumed */
	    if (sfp == NULL) {
	      fprintf(stderr, "Could not use standard input as sequence file input.\n");
	    } else {
	      data.entries = sl_get_sequence_list(sfp, data.entries);
	      seqfclose(sfp);
	    }
	    break;

	case '\0':
	    input_given = 1;
	    sfp = seqfopen(optarg, "r", NULL); /* No format assumed */
	    if (sfp == NULL) {
	      fprintf(stderr, "Could not open '%s' as a sequence file.\n", optarg);
	    } else {
	      data.entries = sl_get_sequence_list(sfp, data.entries);
	      seqfclose(sfp);
	    }	    
	    break;
 
	case 'o':
	    if (!(outfile = freopen(optarg, "w", stdout))) {
		fprintf(stderr, "Couldn't open '%s' for writing.\n", optarg);
		exit(1);
	    }
	    outfilename = strdup(optarg);
	    option++;
	    break;

	case 'k':
	    keep_seqio_prefix = 1;
	    break;
	    
	case 's':
	    shorten_long_seqnames = 1;
	    break;

	case 'v':
	    verbose = 1;
	    break;

	default:
	    errflag = 1;
	}
    }
 
    /*
      If something went wrong when parsing commandline arguments,
      print the usage message.
      */

    if (errflag) {
	fprintf(stderr, "usage: %s [-v] [-k] [-s] [-o <outfile>] [-x <infofile>] <infilenames>\n", argv[0]);
	fprintf(stderr, "   Default is writing to stdout.\n");
	fprintf(stderr, "   -c   Write copyright notices.\n");
	fprintf(stderr, "   -v   Verbose. Gives you extra information on the calculations.\n");
	fprintf(stderr, "   -k   Keep prefixes in sequence identifiers introduces by the SEQIO package.\n        Default is to remove them, but beware that ambigouity among identifiers\n        are not checked.\n");
	fprintf(stderr, "   -s   To make sure that the sequence identifiers are at most 10 characters\n        long, which some programs (e.g. in the PHYLIP suite) demands. Note that\n        no checking on uniqueness among identifiers are made.\n");
	fprintf(stderr, "   -x   Specifies a file to put information about the computation in.\n");
	fprintf(stderr, "        If not given, all extra information goes to stderr.\n");
	fprintf(stderr, "   <infilenames>\n        One or more references to sequence files. The name '--' is\n        special and makes %s read from stdin.\n", argv[0]);
	exit(2);
    }

    if (input_given == 0) {
      fprintf(stderr, "Reading files from stdin.\n");
      sfp = seqfopen("-", "r", NULL); /* No format assumed */
      if (sfp == NULL) {
	fprintf(stderr, "Could not use standard input as sequence file input.\n");
      } else {
	data.entries = sl_get_sequence_list(sfp, data.entries);
	seqfclose(sfp);
      }
    }

    if (print_copyright_notice) {
      fprintf(stderr, "%s\n", lanl_copyright);
      fprintf(stderr, "%s\n", seqio_copyright);
      fprintf(stderr, "%s\n", cephes_copyright);
    }

    if (verbose) {
	current_time = time(NULL);
	rev_string = strdup(strchr(REVISION, ' '));
	fprintf(stderr, "Version:%s, (revision %s)\nJob started %s", version_string, strtok(rev_string,"$"), ctime(&current_time));
	free(rev_string);
	if (outfilename) {
	    fprintf(stderr, "Outfile: %s; ", outfilename);
	}
	if (informationfilename) {
	    fprintf(stderr, "Information file: %s; ", informationfilename);
	}
	fprintf(stderr, "\n");
	if (keep_seqio_prefix == 0) {
	  fprintf(stderr, "Prefix on sequence identifiers are removed.\n");
	}
	if (shorten_long_seqnames) {
	  fprintf(stderr, "Long sequence identifiers are shortened to at most 10 characters.\n");
	}
	fprintf(stderr, "\n");
    }

    if (keep_seqio_prefix == 0) {
      sl_remove_prefix(data.entries, 1);
    }
    if (shorten_long_seqnames) {
      sl_check_name_length(data.entries);
    }

    data.nr_of_entries = sl_list_length(data.entries);

    if (data.nr_of_entries < 2) {	/* Read sequences. Continue if non-null */
	fprintf(stderr, "Only %d sequences read. Aborts.\n", data.nr_of_entries);
	exit(1);
    } else {
	if (verbose) {
	    fprintf(stderr, "%d sequences read.\nNow removing indel and ambigous sites.\n", data.nr_of_entries);
	}
	/* 
	   Prepare for analysis by removing sites with indels and ambigouity
	   codes. Will warn for different sequence lengths.
	 */
	
	r_val = strip_indels_etc(data, allowed_characters, &new_entries);
	if (r_val != R_OK) {
	  if (r_val | R_DIFFERENT_LENGTHS) {
	    fprintf(stderr, "Sequences have different length! Sequences stripped down to the shortest sequence.\n");
	  }
	  if (r_val | R_NO_DATA) {
	    fprintf(stderr, "Programming error! Aborting. (%s:%d)\n", __FILE__, __LINE__);
	    exit(1);
	  }
	  if (r_val | R_OUT_OF_MEMORY) {
	    fprintf(stderr, "Out of memory! Aborting. (%s:%d)\n", __FILE__, __LINE__);
	    exit(1);
	  }
	} 
#ifdef DEBUG_STRIP
	/* If the stripping is being tested. */
	memset(&info, 0, sizeof(SEQINFO));
	sfp = seqfopen("stripped-seqs.gb", "w", "genbank");
	for (e = new_entries; e != NULL; e=e->next) {
	  info.idlist = e->name;
	  seqfwrite(sfp, e->seq, strlen(e->seq), &info);
	}
#endif DEBUG_STRIP

	sl_free_list_and_sequences(data.entries);
	data.entries = new_entries;

	if (verbose) { 
	  fprintf(stderr, "Sequences are %ld bp long. Now analyzing.\n", strlen(data.entries->seq)); 
	} 

	/*
	  We can start analysing.
	  For starters, go through every sequence pair and count the 
	  different substitutions.
	  */

	data.base_count = matrix_init_square(end_of_nucl_codes); 
	/* Matrix used for collecting number of different base occurances */
    
	data.data = calc_pairwise_freqs(data.entries, data.base_count); 
	/* nucl_count stores a return value */
	/* The matrices in the data.data list contains the number
	   of times nucleotide s had an observed transition
	   to nucleotide t in position (s, t).
	 */

	
	if (verbose) {
	    fprintf(stderr, "All sequence pairs analyzed.\n");
	    fprintf(stderr, "Counted nucleotides: \n");
	    matrix_fprint(stderr, data.base_count);
	} 
    
	/* 
	   Here, we have counted all nucleotides. 
	   Compute the interesting stuff.
	 */
	
	if (verbose) {
	    c_err = calc_compute_bruno(&data, &distance_estimates, &error_estimates, &substitution_rate_matrix, informationfile);
	} else {
	    c_err = calc_compute_bruno(&data, &distance_estimates, &error_estimates, &substitution_rate_matrix, NULL);	    
	}

	if (c_err != C_NO_ERROR) {
	    print_calc_error(c_err);
	    exit(1);
	}

	if (verbose) {
	    printf("\n\n");
	}

	name_length=0;
	for (current_entry = data.entries, i = 0; 
	     current_entry != NULL; 
	     current_entry = current_entry->next, i++) {
	  if (name_length < strlen(current_entry->name)) {
	    name_length = strlen(current_entry->name);
	  }
	}
	sprintf(format, "%%-%ds", name_length+1);
	printf("\t%d\n", data.nr_of_entries);
	for (current_entry = data.entries, i = 0; 
	     current_entry != NULL; 
	     current_entry = current_entry->next, i++) {
/* 	    printf("%-11s", current_entry->name); */
	    printf(format, current_entry->name);
	    matrix_fprint_row(stdout, distance_estimates, i);
	    printf("\n");
	}
	
	printf("\nError estimates for the distances:\n");
	for (current_entry = data.entries, i = 0; 
	     current_entry != NULL; 
	     current_entry = current_entry->next, i++) {
	    printf(format, current_entry->name);
	    matrix_fprint_row(stdout, error_estimates, i);
	    printf("\n");
	}
	
	printf("\nSubstitution rate estimate (ACGT):\n");
	matrix_print(substitution_rate_matrix);

    }
	exit(0);			/* Done! */
}
