/*
Copyright, 1994-1995, The Regents of the University of California. 
This software was produced under U.S. Government contract(XXXXX)
by Los Alamos National Laboratory, which is operated by the University 
of California for the U.S. Department of Energy. The U.S Government 
is licensed to use, reproduce, and distribute this software. Permission 
is granted to the public to copy and use this software without charge, 
provided that this Notice and any statement of authorship are reproduced 
on all copies. Neither the Government nor the University makes any 
warranty, express or implied, or assumes any liability or responsibility 
for the use of this software.
*/

/* $Id: matrix.c,v 1.2 96/09/11 13:33:26 arve Exp $ */

#include <math.h>
#include "matrix.h"

/*
 * Allocate memory for distance matrix and fill it with zeroes
 */

float_matrix*
matrix_init(unsigned rows, unsigned columns) {
  float_matrix *tmp;
  float *m;
  unsigned i,j;
 
  if ((tmp = (float_matrix*) malloc(sizeof(float_matrix)))) {
    if ((m = (float*) malloc(rows*columns*sizeof(float)))) {
      tmp->matrix = m;
      tmp->size = ((rows == columns) ? rows : -1);
      tmp->rows = rows;
      tmp->columns = columns;
      for (i=0; i<rows; i++) {
	for (j=0; j<columns; j++) {
	  *(tmp->matrix + (i*columns) + j) = 0;
	}
      }
      tmp->max_val = 0;
      tmp->min_val = 0;
    } else {
      fprintf(stderr, "Not enough memory!\n");
      exit(2);
    }
  } else {
    fprintf(stderr, "Not enough memory!\n");
    exit(2);
  }
  return tmp;
}


float_matrix *
matrix_init_square(unsigned size) {
  return matrix_init(size, size);
}


matrix_errors
matrix_free(float_matrix *m) {
  if (m) {
    free(m->matrix);
    free (m);
    return M_NO_ERROR;
  } else {
    return M_NULL_MATRIX;
  }
}


matrix_errors
matrix_set(float_matrix *m, unsigned row, unsigned col, float val) {
  if (m) {
    if ((col < m->columns) && (row < m->rows)) {
      *(m->matrix + (row*m->columns) + col) = val;
      return M_NO_ERROR;
    } else {
      return M_OUT_OF_BOUNDS;
    } 
  } else {
    return M_NULL_MATRIX;	
  }
}


matrix_errors
matrix_set_all(float_matrix *m, float values[]) {
  int row, col;

  if (m) {
    for (row = 0; row < matrix_rows(m); row++) {
      for (col = 0; col < matrix_columns(m); col++) {
	matrix_set(m, row, col, values[row*matrix_columns(m) + col]);
      }
    }
    return M_NO_ERROR;
  } else {
    return M_NULL_MATRIX;	
  }
}


matrix_errors
matrix_copy(float_matrix *m1, float_matrix *m2) {
  unsigned i, j;

  if (m1 && m2) {
    if ((matrix_rows(m1) == matrix_rows(m2)) &&
	(matrix_columns(m1) == matrix_columns(m2))) {
      for (i = 0; i < matrix_rows(m1); i++) {
	for (j = 0; j < matrix_columns(m1); j++) {
	  matrix_set(m2, i, j, matrix_val(m1, i, j));
	}
      }
      return M_NO_ERROR;
    } else {
      return M_WRONG_SIZE_OF_MATRIX; 
    } 
  } else {
    return M_NULL_MATRIX;
  }
}


matrix_errors
matrix_copy_column(float_matrix *from, unsigned from_col, 
		   float_matrix *to, unsigned to_col) {
  unsigned row;

  if (from && to) {
    if ((from_col < matrix_columns(from))
	&& (to_col < matrix_columns(to))) {
      if (matrix_rows(from) == matrix_rows(to)) {
	for (row = 0; row < matrix_rows(from); row++) {
	  matrix_set(to, row, to_col, matrix_val(from, row, from_col));
	}
	return M_NO_ERROR;
      } else {
	return M_WRONG_SIZE_OF_MATRIX;
      }
    } else {
      return M_OUT_OF_BOUNDS;	
    }      
  } else {
    return M_NULL_MATRIX;			
  }
}


matrix_errors
matrix_copy_row(float_matrix *from, unsigned from_row, 
		float_matrix *to, unsigned to_row) {
  unsigned col;
  
  if (from && to) {
    if ((from_row < matrix_rows(from))
	&& (to_row < matrix_rows(to))) {
      if (matrix_columns(from) == matrix_columns(to)) {
	for (col = 0; col < matrix_columns(from); col++) {
	  matrix_set(to, to_row, col, matrix_val(from, from_row, col));
	}
	return M_NO_ERROR;
      } else {
	return M_WRONG_SIZE_OF_MATRIX;
      }
    } else {
      return M_OUT_OF_BOUNDS;	
    }
  } else {
    return M_NULL_MATRIX;			
  }
}
      


matrix_errors
matrix_transpose(float_matrix *m) {
  unsigned i, j;
  float tmp;

  if (m) {
    if (matrix_rows(m) == matrix_columns(m)) {
      for (i = 0; i < matrix_rows(m); i++) {
	for (j = i+1; j < matrix_columns(m); j++) {
	  tmp = matrix_val(m, i, j);
	  matrix_set(m, i, j, matrix_val(m, j, i));
	  matrix_set(m, j, i, tmp);
	}
      }
      return 0;
    } else {
      fprintf(stderr, "Tried transpose non-square matrix. This is not implemented!\n");
      exit(1);
    }
  } else {
    return 1;			/* No matrix */
  }
}
	    

float_matrix *
matrix_make_transpose(float_matrix *m) {
  float_matrix *new;
  int rows, columns;
  int i, j;
  matrix_errors status;

  if (m) {
    rows = matrix_rows(m);
    columns = matrix_columns(m);
    new = matrix_init(columns, rows);
    if (new) {
      for (i = 0; i < rows; i++) {
	for (j = 0; j < columns; j++) {
	  status = matrix_set(new, j, i, matrix_val(m, i, j));
	  if (status != M_NO_ERROR) {
	    print_matrix_error(status);
	    exit(1);
	  }
	}
      }
      return new;
    } else {
      return NULL;
    }
  } else {
    return NULL;
  }
} 



matrix_errors 
matrix_normalize_vector(float_matrix *v) {
  float norm=0;
  unsigned i;

  if (!v) {
    return M_NULL_MATRIX;
  } else {
    if (matrix_rows(v) == 1) {
      for (i = 0; i < matrix_columns(v); i++) {
	norm += matrix_val(v, 0, i)*matrix_val(v, 0, i);
      }
      norm = sqrt(norm);
      for (i = 0; i < matrix_columns(v); i++) {
	matrix_set(v, 0, i, matrix_val(v, 0, i) / norm);
      }	
      return M_NO_ERROR;

    } else if (matrix_columns(v) == 1) {
      for (i = 0; i < matrix_rows(v); i++) {
	norm += matrix_val(v, i, 0)*matrix_val(v, i, 0);
      }
      norm = sqrt(norm);
      for (i = 0; i < matrix_rows(v); i++) {
	matrix_set(v, i, 0, matrix_val(v, i, 0) / norm);
      }
      return M_NO_ERROR;

    } else {
      return M_WRONG_SIZE_OF_MATRIX;
    }
  } 
}



matrix_errors
matrix_normalize_column(float_matrix *m, unsigned i) {
  float_matrix *tmp;
  matrix_errors status;

  if (m) {
    if (i < matrix_columns(m)) {
      tmp = matrix_init(4, 1);
      status = matrix_copy_column(m, i, tmp, 0);
      if (status != M_NO_ERROR) {
	return status;
      }
      status = matrix_normalize_vector(tmp);
      if (status != M_NO_ERROR) {
	return status;
      }
      matrix_copy_column(tmp, 0, m, i);
      return M_NO_ERROR;
    } else {
      return M_OUT_OF_BOUNDS;
    }
  } else {
    return M_NULL_MATRIX;
  }
}
      


float
matrix_val(float_matrix *m, unsigned row, unsigned col) {
  if ((col < m->columns) && (row < m->rows)) {
    return *(m->matrix + (row*m->columns) + col);
  } else {
    fprintf(stderr, "Programming error! Tried to read out of bounds! (Row %d, col %d on matrix of size %d x %d)\n", row, col, m->rows, m->columns);
    exit(2);
  }
}


float
matrix_max(float_matrix *m) {
  if (m) {
    return (m->max_val);
  } else {
    fprintf(stderr, "Programming error: Tried to work with empty distance matrix.\n");
    exit(1);
  }
}


float
matrix_min(float_matrix *m) {
  if (m) {
    return (m->min_val);
  } else {
    fprintf(stderr, "Programming error: Tried to work with empty distance matrix.\n");
    exit(1);
  }
}



float
matrix_val_span(float_matrix *m) {
  if (m) {
    return (m->max_val - m->min_val);
  } else {
    fprintf(stderr, "Programming error: Tried to work with empty distance matrix.\n");
    exit(1);
  }
}


int
matrix_size(float_matrix *m) {
  if (m) {
    return (m->size);
  } else {
    return -1;
  }
}

int
matrix_rows(float_matrix *m) {
  if (m) {
    return (m->rows);
  } else {
    return -1;
  }
}

int
matrix_columns(float_matrix *m) {
  if (m) {
    return (m->columns);
  } else {
    return -1;
  }
}



matrix_errors
matrix_set_max(float_matrix *m, float max) {
  if (m) {
    m->max_val = max;
    return M_NO_ERROR;
  } else {
    return M_NULL_MATRIX;
  }
}





matrix_errors
matrix_set_min(float_matrix *m, float min) {
  if (m) {
    m->min_val = min;
    return M_NO_ERROR;
  } else {
    return M_NULL_MATRIX;
  }
}

matrix_errors
matrix_add_to(float_matrix *a, float_matrix *sum) {
  int i, j;
  
  if (a && sum) {
    if ((matrix_rows(a) == matrix_rows(sum)) &&
	(matrix_columns(a) == matrix_columns(sum))) {
      for (i = 0; i < matrix_rows(a); i++) {
	for (j = 0; j < matrix_columns(a); j++) {
	  matrix_set(sum, i, j, matrix_val(a, i, j) + matrix_val(sum, i, j));
	}
      }
      return M_NO_ERROR;
    } else {
      return M_WRONG_SIZE_OF_MATRIX;
    }
  } else {
    return M_NULL_MATRIX;
  }
}


matrix_errors
matrix_mul(float_matrix *a, float_matrix *b, float_matrix *res) {
  unsigned i, j, k;
  float tmp;

  if ((a && b) && res) {
    if ((matrix_columns(a) == matrix_rows(b)) &&
	(matrix_rows(a) == matrix_rows(res)) &&
	(matrix_columns(b) == matrix_columns(res))) {
      for (i = 0; i < matrix_rows(a); i++) {
	for (j = 0; j < matrix_columns(b); j++) {
	  tmp = 0;
	  for (k = 0; k < matrix_rows(b); k++) {
	    tmp += matrix_val(a, i, k) * matrix_val(b, k, j);
	  }
	  matrix_set(res, i, j, tmp);
	}
      }
      return M_NO_ERROR;
    } else {			/* Can't multiply */
      return M_WRONG_SIZE_OF_MATRIX;
    }
  } else {			/* Matrix missing */
    return M_NULL_MATRIX;
  }    
}


matrix_errors
matrix_scalar_mul(float_matrix *m, float s) {
  int row, col;

  if (m) {
    for (row = 0; row < matrix_rows(m); row++) {
      for (col = 0; col < matrix_columns(m); col++) {
	matrix_set(m, row, col, 
		   s * matrix_val(m, row, col));
      }
    }
    return M_NO_ERROR;
  } else {
    return M_NULL_MATRIX;
  }
}

/*
  matrix_eigen
  This function is depending on the 'eigen' function in the Cephes package.
  Here, we just preper for a call to eigen and offer a nicer interface.
*/

matrix_errors
matrix_eigen(float_matrix *m, float_matrix *eigenvects, float_matrix *eigenvals) {
  int row, col;
  int n;
  double *A, *EV, *E;

  if ((m && eigenvects) && eigenvals) {
    n = matrix_size(m);
    if ((n == matrix_size(eigenvects) 
	 && (n == matrix_rows(eigenvals))
	 && (matrix_columns(eigenvals) == 1))) {
      if (n > 0) {
	/* Prepare variables for call to the cephes function eigens */
	A = (double *) malloc(n*(n+1)*sizeof(double)/2);
	EV = (double *) malloc(n*n*sizeof(double));
	E = (double *) malloc(n*sizeof(double));
	if ((A && EV) && E) {
#ifdef MATRIX_DEBUG
	      printf("Debug info from eigenvects() in matrix.c:\n");
#endif
	  for (row = 0; row < n; row++) {
	    for (col = 0; col <= row; col++) {
	      A[(row*row+row)/2 + col] = (double) matrix_val(m, row, col);
#ifdef MATRIX_DEBUG
	      printf("%f\t", A[(row*row+row)/2 + col]);
#endif
	    }
#ifdef MATRIX_DEBUG
	      printf("\n");
#endif
	  }

	  /* Now, call the 'foreign' function and trust that no error occurs.*/
	  eigens(A, EV, E, n);
	  
	  /*  Write result in our format. */
	  for (row = 0; row < n; row++) {
	    for (col = 0; col < n; col++) {
	      matrix_set(eigenvects, row, col, (float) EV[col*n + row]);
	    }
	  }
	  for (row = 0; row < n; row++) {
	    matrix_set(eigenvals, row, 0, (float) E[row]);
	  }
	  
	  free(A);
	  free(EV);
	  free(E);
	  
	  return M_NO_ERROR;
	} else {
	  return M_OUT_OF_MEMORY; /* Out of memory */
	}
      } else {
	return M_WRONG_SIZE_OF_MATRIX; /* Not a square matrix */
      }
    } else {
      return M_WRONG_SIZE_OF_MATRIX; /* Wrong sizes on result matrix and vector
*/
    }
  } else {
    return M_NULL_MATRIX;	/* Null matrices supplied */
  }
}


/*
  matrix_inverse
  For the moment, this function can only invert 3x3 matrices... 
  But that is all I need right now!

  You get the inverse delivered in the second argument, a matrix that you 
  must supply initialized.
*/

matrix_errors
matrix_inverse(float_matrix *m, float_matrix *inverse) {
  double determinant;
  double m11, m12, m13, m21, m22, m23, m31, m32, m33;
  
  if (!m || !inverse) {
    return M_NULL_MATRIX;
  }

  if (matrix_size(m) != 3) {
    return M_WRONG_SIZE_OF_MATRIX;
  }

  m11=matrix_val(m, 0, 0); m12=matrix_val(m, 0, 1); m13=matrix_val(m, 0, 2);
  m21=matrix_val(m, 1, 0); m22=matrix_val(m, 1, 1); m23=matrix_val(m, 1, 2);
  m31=matrix_val(m, 2, 0); m32=matrix_val(m, 2, 1); m33=matrix_val(m, 2, 2); 

  determinant = m11*m22*m33 + m12*m23*m31 + m13*m21*m32 
      - m13*m22*m31 - m12*m21*m33 - m11*m23*m32;
  
  matrix_set(inverse, 0, 0, m22*m33 - m23*m32);
  matrix_set(inverse, 0, 1,-m12*m33 + m13*m32);
  matrix_set(inverse, 0, 2, m12*m23 - m13*m22);

  matrix_set(inverse, 1, 0,-m21*m33 + m23*m31);
  matrix_set(inverse, 1, 1, m11*m33 - m13*m31);
  matrix_set(inverse, 1, 2,-m11*m23 + m13*m21);

  matrix_set(inverse, 2, 0, m21*m32 - m22*m31);
  matrix_set(inverse, 2, 1,-m11*m32 + m12*m31);
  matrix_set(inverse, 2, 2, m11*m22 - m12*m21);

  matrix_scalar_mul(inverse, 1.0/determinant);

  return M_NO_ERROR;
} /* matrix_inverse */



matrix_errors
matrix_fprint_row(FILE *f, float_matrix *m, int i) {
  int j;

  if (m) {
    if ((i >= 0) && (i < matrix_rows(m))) {
      for (j = 0; j < matrix_columns(m); j++) {
	fprintf(f, "%f\t", matrix_val(m, i, j));
      }
    } else {
      return M_OUT_OF_BOUNDS;	/* Row number impossible */
    }
  } else {
    return M_NULL_MATRIX;	/* No matrix supplied */
  }
  return M_NO_ERROR;
}

matrix_errors
matrix_fprint(FILE *f, float_matrix *m) {
  int i;
  matrix_errors status;
  
  if (m) {
    for (i = 0; i < matrix_rows(m); i++) {
      status = matrix_fprint_row(f, m, i);
      if (status != M_NO_ERROR) {
	return status;		/* Error, could not print row*/
      } else {
	fprintf(f, "\n");
      }
    }
    return 0;
  } else {
    return 1;			/* Error, no matrix */
  }
}


float_matrix*
matrix_read(FILE *f, unsigned rows, unsigned columns){
  float_matrix *m;
  unsigned r, c;
  float in;

  m = matrix_init(rows, columns);

  if (m) {
    for (r = 0; r < rows; r++) {
      for (c = 0; c < columns; c++) {
	if (1 == fscanf(f, "%f", &in)) {
	  matrix_set(m, r, c, in);
	} else {
	  fprintf(stderr, "Something is wrong with the input for matrix read from file.\nCheck your format!\n");
	  exit(1);
	}
      }
    }
  } else {
    fprintf(stderr, "Out of memory when trying to create matrix in functon 'matrix_read'\n");
    exit(1);
  }
  return m;
}


char *matrix_error_codes[] = {
  "No error.",
  "Out of memory.",
  "Null matrix to matrix function.",
  "Out of bounds on matrix.",
  "Wrong size of matrix.",
  ""
};

void
print_matrix_error(matrix_errors errflag) {
  if (errflag < M_END_OF_ERRORS) {
    fprintf(stderr, "\n%s\n", matrix_error_codes[errflag]);
  } else {
    fprintf(stderr, "\nUnknown matrix related error. Don't panic.\n");
  }
}
